IMPLEMENTATION MODULE Scnr;

FROM SYSTEM IMPORT ADR;
IMPORT Lib, Str, IO, FIO, Rdr, Ints;


TYPE
  TStrIdx = [ 0 .. 15 ];
  TString = ARRAY TStrIdx OF CHAR;

VAR
  chr: Rdr.TSymbol;
  pos: TPosition;
  val: Ints.TValue;


PROCEDURE Get();
BEGIN
  chr := Rdr.Read();
END Get;


PROCEDURE Init( s-: ARRAY OF CHAR );
BEGIN
  Rdr.Init( s );
  Get();
  pos := 0;
END Init;


PROCEDURE Error( p: TPosition; s-: ARRAY OF CHAR );
VAR
  i: TPosition;
BEGIN
  IO.WrStr( '(' );
  IO.WrCard( p, 1 );
  IO.WrStr( ') Scanner error: ' );
  IO.WrStr( s );
  IO.WrStr( '.' );
  IO.WrLn();
  Rdr.OutStr();
  (* show position -- function requires tab to be an invalid char ! *)
  FOR i := 1 TO p DO IO.WrChar( '_' ) END;
  IO.WrChar( '^' );
  IO.WrLn();
  HALT;
END Error;


(* sym = num | "(" | ")" | "+" | "-" | "*" | "/" .
 * num = dig ( { dig } | { hdg } "H" ) .
 * dig = "0" | "1" | ... | "9" .
 * hdg = dig | "A" | ... | "F" .
 *
 * We must read-ahead one input symbol (chr) since the end of symNum
 * is detectable only with the next input symbol.
 *)
VAR
  str: TString;

PROCEDURE Read(): TSymbol;

  PROCEDURE IsHexDigit(): BOOLEAN;
  BEGIN
    RETURN ( ( chr >= '0' ) & ( chr <= '9' ) ) OR
           ( ( chr >= 'A' ) & ( chr <= 'F' ) ) OR
           ( chr = 'H' );
  END IsHexDigit;

  PROCEDURE Number();
  VAR i: CARDINAL;

    PROCEDURE Put();
    BEGIN
      IF i > MAX(TStrIdx) THEN Error( Rdr.Pos(), 'Too many digits' ) END;
      str[ i ] := chr;
      INC( i );
      Get();
    END Put;

  VAR
    b: CARDINAL;
    ok: BOOLEAN;
  BEGIN
    Lib.Fill( ADR(str), SIZE(str), 0 );
    i := MIN(TStrIdx);
    REPEAT Put() UNTIL ~ IsHexDigit();
    CASE str[i-1] OF
    | 'H': b := 16; str[i-1] := 0C;
    | 'C': b :=  8; str[i-1] := 0C;
    | 'B': b :=  2; str[i-1] := 0C;  (* Wirth: should be b := 8 *)
    ELSE
      b := 10;
    END;
    Ints.StrToVal( val, str, b, ok );
    IF ~ ok THEN
      CASE b OF
      | 16: Error( pos, 'Invalid hexadecimal number' );
      | 10: Error( pos ,'Invalid decimal number' );
      |  8: Error( pos, 'Invalid octal number' );
      |  2: Error( pos, 'Invalid binary number' );
      END;
    END;
  END Number;

VAR
  sym: TSymbol;
BEGIN
  WHILE chr = ' ' DO Get() END;
  pos := Rdr.Pos();

  IF chr = Rdr.symEOS THEN
    val := Ints.valZero;
    RETURN symEOS;
  END;

  CASE chr OF
  | '0' .. '9': sym := symNum; Number();
  | '('       : sym := symLBrk; Get();
  | ')'       : sym := symRBrk; Get();
  | '+'       : sym := symAdd; Get();
  | '-'       : sym := symSub; Get();
  | '*'       : sym := symMul; Get();
  | '/'       : sym := symDiv; Get();
  ELSE
    Error( Rdr.Pos(), 'Unexpected character' )
  END;
  RETURN sym;
END Read;


PROCEDURE Val(): TValue;
BEGIN
  RETURN val;
END Val;

PROCEDURE Pos(): TPosition;
BEGIN
  RETURN pos;
END Pos;

PROCEDURE OutStr();
BEGIN
  Rdr.OutStr();
END OutStr;

BEGIN
  chr := Rdr.symEOS;
  pos := 0;
  val := Ints.valZero;
END Scnr.
